#include <stdio.h>
#include <string.h>
#include <immintrin.h>

#include "curve25519_basic.h"
#include "crypto_scalarmult.h"

#define force_inline __attribute__((always_inline)) inline
#define a24 121665

/*
    For each packed 64-bit element:
    Multiply the 64 bit number in a with the 32 bit number in b
    returning 64 bit. Useful if a*b < 2^64
    latency: 7

    For larger results, i.e. a*b < 2^128 or a*b < 2^96, the latency is at least 8 and only 2 such multiplications can be vectorized.
    Because of that it's not worth it to vectorize.
*/
force_inline
__m256i mul64x32_low (__m256i a, __m256i b) {

    // __m256i prodh = _mm256_mullo_epi32(a, b);           // xxx, a0H*b, xxx, a1H*b
    // __m256i prodh2 = _mm256_and_si256(prodh, _mm256_set1_epi64x(0xFFFFFFFF00000000)); // zero low halves
    // latency 8 + 1 = 9 (mullo is expensive!)

    //__m256i a_h = _mm256_srli_epi64(a, 32);
    __m256i a_h = _mm256_shuffle_epi32(a, 0xB1);                // xxx, a0H, xxx, a1H
    __m256i prodh = _mm256_mul_epi32(a_h, b);                   // a0h * b, a1H * b, (64 bit products) prodh < 2^32
    __m256i prodh2 = _mm256_slli_epi64(prodh, 32);

    __m256i prodl  = _mm256_mul_epu32(a, b);               // a0L*b, a1L*b, (64 bit unsigned products)

    __m256i prod    = _mm256_add_epi64(prodl,prodh2);
    return  prod;
}

/*
    rL = aL + bL
    rH = aH + bH + (rL < bL)
    schedulling for each cycle:
    0: add (port 0), add (port 1)
    1: xor (port 0), xor (port 1)
    2: cmp (port 0)
    3: cmp
    4: cmp
    5: sub (port 0)

    This occupies over its execution 8 times a port.
    For each 128-bit number this are 2 ports.
    Same as when add and adc are used on a 128-bit number as scalar.
    Hence, vectorization does not clearly improve performance.
*/
void add128 (__m256i *rH, __m256i *rL, __m256i aH, __m256i aL, __m256i bH, __m256i bL){
    __m256i sign64 = _mm256_set1_epi64x(0x8000000000000000L);

    __m256i resH = _mm256_add_epi64(aH, bH);

    __m256i resL = _mm256_add_epi64(aL, bL);

    // resL < bL (unsigned)
    __m256i resLU = _mm256_xor_si256(resL, sign64);
    __m256i bLU = _mm256_xor_si256(bL, sign64);
    __m256i cmp = _mm256_cmpgt_epi64(resLU, bLU); // -(resL < bL)

    *rH = _mm256_sub_epi64(resH, cmp);
    *rL = resL;
}

force_inline
void field_add(field_t dst, field_t a, field_t b) {
    __m256i a_0 = _mm256_load_si256((__m256i*) a);
    __m256i b_0 = _mm256_load_si256((__m256i*) b);

    __m256i dst_0 = _mm256_add_epi64(a_0, b_0);

    _mm256_store_si256 ((__m256i*) dst, dst_0);
    
    dst[4] = a[4] + b[4];
}

force_inline
void field_sub(field_t dst, field_t a, field_t b) {
    __m256i a_0 = _mm256_load_si256((__m256i*) a);
    __m256i b_0 = _mm256_load_si256((__m256i*) b);

    __m256i dst_0 = _mm256_sub_epi64(a_0, b_0);

    _mm256_store_si256 ((__m256i*) dst, dst_0);

    dst[4] = a[4] - b[4];
}

// dst0 = a + b
// dst1 = a - b
force_inline
void field_add_sub(field_t dst0, field_t dst1, field_t a, field_t b) {
    __m256i a_0 = _mm256_load_si256((__m256i*) a);    // 0
    __m256i b_0 = _mm256_load_si256((__m256i*) b);    // 0

    __m256i dst0_0 = _mm256_add_epi64(a_0, b_0);  // 1
    __m256i dst1_0 = _mm256_sub_epi64(a_0, b_0);  // 1

    _mm256_store_si256 ((__m256i*) dst0, dst0_0);     // 0
    _mm256_store_si256 ((__m256i*) dst1, dst1_0);     // 0

    // TODO: combine in one vector?
    dst0[4] = a[4] + b[4];
    dst1[4] = a[4] - b[4];
}

force_inline
void field_carry(int128_t a[5]) {
    int128_t carry[5];
    
    carry[0] = (a[0] + ((int64_t)1 << 50)) >> 51;
    a[1] += carry[0];
    a[0] -= carry[0] << 51;

    carry[1] = (a[1] + ((int64_t)1 << 50)) >> 51;
    a[2] += carry[1];
    a[1] -= carry[1] << 51;

    carry[2] = (a[2] + ((int64_t)1 << 50)) >> 51;
    a[3] += carry[2];
    a[2] -= carry[2] << 51;

    carry[3] = (a[3] + ((int64_t)1 << 50)) >> 51;
    a[4] += carry[3];
    a[3] -= carry[3] << 51;

    carry[4] = (a[4] + ((int64_t)1 << 50)) >> 51;
    a[0] += carry[4] * 19;
    a[4] -= carry[4] << 51;

    carry[0] = (a[0] + ((int64_t)1 << 50)) >> 51;
    a[1] += carry[0];
    a[0] -= carry[0] << 51;
}

// source: https://github.com/golang/crypto/blob/master/curve25519/curve25519_generic.go
void field_square(field_t dst, field_t src) {
    field_mul(dst, src, src);
}


// source: https://github.com/golang/crypto/blob/master/curve25519/curve25519_generic.go
void field_mul(field_t dst, field_t a, field_t b) {
    int64_t f0 = a[0];
    int64_t f1 = a[1];
    int64_t f2 = a[2];
    int64_t f3 = a[3];
    int64_t f4 = a[4];
    
    int64_t g0 = b[0];
    int64_t g1 = b[1];
    int64_t g2 = b[2];
    int64_t g3 = b[3];
    int64_t g4 = b[4];

    int64_t g1_19 = 19 * g1; // g < 2^51 & 19 < 2^6 => 19*g < 2^57
    int64_t g2_19 = 19 * g2;
    int64_t g3_19 = 19 * g3;
    int64_t g4_19 = 19 * g4;
    
    int128_t f0g0 = (int128_t) f0 * g0;
    int128_t f0g1 =(int128_t) f0 * g1;

    int128_t f0g2 =(int128_t) f0 * g2;
    int128_t f0g3 =(int128_t) f0 * g3;

    int128_t f0g4 =(int128_t) f0 * g4;


    int128_t f1g0 = (int128_t) f1 * g0;
    int128_t f1g1 = (int128_t) f1 * g1;

    int128_t f1g2 = (int128_t) f1 * g2;
    int128_t f1g3 = (int128_t) f1 * g3;

    int128_t f1g4_19 = (int128_t) f1 * g4_19;


    int128_t f2g0 = (int128_t) f2 * g0;
    int128_t f2g1 = (int128_t) f2 * g1;

    int128_t f2g2 = (int128_t) f2 * g2;

    int128_t f2g3_19 = (int128_t) f2 * g3_19;
    int128_t f2g4_19 = (int128_t) f2 * g4_19;
    
    int128_t f3g0 = (int128_t) f3 * g0;
    int128_t f3g1 = (int128_t) f3 * g1;

    int128_t f3g2_19 = (int128_t) f3 * g2_19;
    int128_t f3g3_19 = (int128_t) f3 * g3_19;

    int128_t f3g4_19 = (int128_t) f3 * g4_19;

    int128_t f4g0 = (int128_t) f4 * g0;

    int128_t f4g1_19 = (int128_t) f4 * g1_19;
    int128_t f4g2_19 = (int128_t) f4 * g2_19;

    int128_t f4g3_19 = (int128_t) f4 * g3_19;
    int128_t f4g4_19 = (int128_t) f4 * g4_19;
    
    int128_t h[5];

    h[0] = f0g0 + f1g4_19 + f2g3_19 + f3g2_19 + f4g1_19;
    h[1] = f0g1 + f1g0    + f2g4_19 + f3g3_19 + f4g2_19;
    h[2] = f0g2 + f1g1    + f2g0    + f3g4_19 + f4g3_19;
    h[3] = f0g3 + f1g2    + f2g1    + f3g0    + f4g4_19;
    h[4] = f0g4 + f1g3    + f2g2    + f3g1    + f4g0;

    field_carry(h);

    dst[0] = (int64_t) h[0];
    dst[1] = (int64_t) h[1];
    dst[2] = (int64_t) h[2];
    dst[3] = (int64_t) h[3];
    dst[4] = (int64_t) h[4];
}


// taken from djb's original implementation
void field_invert(field_t dst, field_t src) {
    field_t z2, z9, z11, z2_5_0, z2_10_0, z2_20_0, z2_50_0, z2_100_0, t0, t1;
    int i;

    field_square(z2,src);           // 2
    field_square(t1,z2);            // 4
    field_square(t0,t1);            // 8
    field_mul(z9,t0,src);           // 9
    field_mul(z11,z9,z2);           // 11
    field_square(t0,z11);           // 22
    field_mul(z2_5_0,t0,z9);        // 2^5 - 2^0 = 31

    field_square(t0,z2_5_0);        // 2^6 - 2^1
    field_square(t1,t0);            // 2^7 - 2^2
    field_square(t0,t1);            // 2^8 - 2^3
    field_square(t1,t0);            // 2^9 - 2^4
    field_square(t0,t1);            // 2^10 - 2^5
    field_mul(z2_10_0,t0,z2_5_0);   // 2^10 - 2^0

    field_square(t0,z2_10_0);       // 2^11 - 2^1
    field_square(t1,t0);            // 2^12 - 2^2
    for (i = 2;i < 10;i += 2) {     // 2^20 - 2^10
        field_square(t0,t1);
        field_square(t1,t0);
    }
    field_mul(z2_20_0,t1,z2_10_0);  // 2^20 - 2^0

    field_square(t0,z2_20_0);       // 2^21 - 2^1
    field_square(t1,t0);            // 2^22 - 2^2
    for (i = 2;i < 20;i += 2) {     // 2^40 - 2^20
        field_square(t0,t1);
        field_square(t1,t0);
    }
    field_mul(t0,t1,z2_20_0);       // 2^40 - 2^0

    field_square(t1,t0);            // 2^41 - 2^1
    field_square(t0,t1);            // 2^42 - 2^2
    for (i = 2;i < 10;i += 2) {     // 2^50 - 2^10
        field_square(t1,t0);
        field_square(t0,t1);
    }
    field_mul(z2_50_0,t0,z2_10_0);  // 2^50 - 2^0

    field_square(t0,z2_50_0);       // 2^51 - 2^1
    field_square(t1,t0);            // 2^52 - 2^2
    for (i = 2;i < 50;i += 2) {     // 2^100 - 2^50
        field_square(t0,t1);
        field_square(t1,t0);
    }
    field_mul(z2_100_0,t1,z2_50_0); // 2^100 - 2^0

    field_square(t1,z2_100_0);      // 2^101 - 2^1
    field_square(t0,t1);            // 2^102 - 2^2
    for (i = 2;i < 100;i += 2) {    // 2^200 - 2^100
        field_square(t1,t0);
        field_square(t0,t1);
    }
    field_mul(t1,t0,z2_100_0);      // 2^200 - 2^0

    field_square(t0,t1);            // 2^201 - 2^1
    field_square(t1,t0);            // 2^202 - 2^2
    for (i = 2;i < 50;i += 2) {     // 2^250 - 2^50
        field_square(t0,t1);
        field_square(t1,t0);
    }
    field_mul(t0,t1,z2_50_0);       // 2^250 - 2^0

    field_square(t1,t0);            // 2^251 - 2^1
    field_square(t0,t1);            // 2^252 - 2^2
    field_square(t1,t0);            // 2^253 - 2^3
    field_square(t0,t1);            // 2^254 - 2^4
    field_square(t1,t0);            // 2^255 - 2^5
    field_mul(dst,t1,z11);          // 2^255 - 21
}

// source: https://github.com/golang/crypto/blob/master/curve25519/curve25519_generic.go

force_inline
void field_mul_a24(field_t dst, field_t src) {
    int128_t h[5];

    h[0] = (int128_t) src[0] * a24;
    h[1] = (int128_t) src[1] * a24;
    h[2] = (int128_t) src[2] * a24;
    h[3] = (int128_t) src[3] * a24;
    h[4] = (int128_t) src[4] * a24;

    field_carry(h);

    dst[0] = (int64_t) h[0];
    dst[1] = (int64_t) h[1];
    dst[2] = (int64_t) h[2];
    dst[3] = (int64_t) h[3];
    dst[4] = (int64_t) h[4];
}

// source: https://github.com/golang/crypto/blob/master/curve25519/curve25519_generic.go
// see bernstein page 5, upper paragrpah for secret key definition
void decode_scalar(byte32_t dst, byte32_t src) {
    memcpy(dst, src, sizeof(byte32_t));
    dst[0] &= 248; // ensure multiple of 8 (248 = 1111 1000)
    dst[31] &= 127; // ensure biggest secret key <= 8*(2^251 - 1) = 2^254 - 1 (127 = 0111 1111)
    dst[31] |= 64; // ensure secret key offset of + 2^254 (64 = 0100 0000) 
}


// source: supercop amd64-51
void decode_point(field_t dst, byte32_t src) {
    int64_t h[5];

    h[0] = (int64_t) src[0];
    h[0] += (int64_t) src[1] << 8;
    h[0] += (int64_t) src[2] << 16;
    h[0] += (int64_t) src[3] << 24;
    h[0] += (int64_t) src[4] << 32;
    h[0] += (int64_t) src[5] << 40;
    h[0] += ((int64_t) src[6] & 7) << 48; // take the first 3 bits

    h[1] = (int64_t) src[6] >> 3;         //skip the first 3 bits (take 5 bits)
    h[1] += (int64_t) src[7] << 5;
    h[1] += (int64_t) src[8] << 13;
    h[1] += (int64_t) src[9] << 21;
    h[1] += (int64_t) src[10] << 29;
    h[1] += (int64_t) src[11] << 37;
    h[1] += ((int64_t) src[12] & 63) << 45;

    h[2] = (int64_t) src[12] >> 6;
    h[2] += (int64_t) src[13] << 2;
    h[2] += (int64_t) src[14] << 10;
    h[2] += (int64_t) src[15] << 18;
    h[2] += (int64_t) src[16] << 26;
    h[2] += (int64_t) src[17] << 34;
    h[2] += (int64_t) src[18] << 42;
    h[2] += ((int64_t) src[19] & 1) << 50;

    h[3] = (int64_t) src[19] >> 1;
    h[3] += (int64_t) src[20] << 7;
    h[3] += (int64_t) src[21] << 15;
    h[3] += (int64_t) src[22] << 23;
    h[3] += (int64_t) src[23] << 31;
    h[3] += (int64_t) src[24] << 39;
    h[3] += ((int64_t) src[25] & 15) << 47;

    h[4] = (int64_t) src[25] >> 4;
    h[4] += (int64_t) src[26] << 4;
    h[4] += (int64_t) src[27] << 12;
    h[4] += (int64_t) src[28] << 20;
    h[4] += (int64_t) src[29] << 28;
    h[4] += (int64_t) src[30] << 36;
    h[4] += ((int64_t) src[31] & 127) << 44;

    //field_carry(h);

    dst[0] = (int64_t) h[0];
    dst[1] = (int64_t) h[1];
    dst[2] = (int64_t) h[2];
    dst[3] = (int64_t) h[3];
    dst[4] = (int64_t) h[4];

}


// source: supercop amd64-51
void encode_point(byte32_t dst, field_t src) {
    int128_t carry[5];

    int64_t q = (19*src[4] + ((int64_t)1 << 50)) >> 51;
    q = (src[0] + q) >> 51;
    q = (src[1] + q) >> 51;
    q = (src[2] + q) >> 51;
    q = (src[3] + q) >> 51;
    q = (src[4] + q) >> 51;

    // Goal: Output h-(2^255-19)q, which is between 0 and 2^255-20.
    src[0] += 19 * q;
    // Goal: Output h-2^255 q, which is between 0 and 2^255-20.

    carry[0] = src[0] >> 51;
    src[1] += carry[0];
    src[0] -= carry[0] << 51;
    carry[1] = src[1] >> 51;
    src[2] += carry[1];
    src[1] -= carry[1] << 51;
    carry[2] = src[2] >> 51;
    src[3] += carry[2];
    src[2] -= carry[2] << 51;
    carry[3] = src[3] >> 51;
    src[4] += carry[3];
    src[3] -= carry[3] << 51;
    carry[4] = src[4] >> 51;
    //src[5] += carry[4];
    src[4] -= carry[4] << 51;



    dst[0] = (uint8_t) (src[0] >> 0);
    dst[1] = (uint8_t) (src[0] >> 8);
    dst[2] = (uint8_t) (src[0] >> 16);
    dst[3] = (uint8_t) (src[0] >> 24);
    dst[4] = (uint8_t) (src[0] >> 32);
    dst[5] = (uint8_t) (src[0] >> 40);

    dst[6] = (uint8_t) (src[0] >> 48) | (src[1] << 3);

    dst[7] = (uint8_t) (src[1] >> 5);
    dst[8] = (uint8_t) (src[1] >> 13);
    dst[9] = (uint8_t) (src[1] >> 21);
    dst[10] = (uint8_t) (src[1] >> 29);
    dst[11] = (uint8_t) (src[1] >> 37);

    dst[12] = (uint8_t) (src[1] >> 45) | (src[2] << 6);

    dst[13] = (uint8_t) (src[2] >> 2);
    dst[14] = (uint8_t) (src[2] >> 10);
    dst[15] = (uint8_t) (src[2] >> 18);
    dst[16] = (uint8_t) (src[2] >> 26);
    dst[17] = (uint8_t) (src[2] >> 34);
    dst[18] = (uint8_t) (src[2] >> 42);

    dst[19] = (uint8_t) (src[2] >> 50) | (src[3] << 1);

    dst[20] = (uint8_t) (src[3] >> 7);
    dst[21] = (uint8_t) (src[3] >> 15);
    dst[22] = (uint8_t) (src[3] >> 23);
    dst[23] = (uint8_t) (src[3] >> 31);
    dst[24] = (uint8_t) (src[3] >> 39);

    dst[25] = (uint8_t) (src[3] >> 47) | (src[4] << 4);

    dst[26] = (uint8_t) (src[4] >> 4);
    dst[27] = (uint8_t) (src[4] >> 12);
    dst[28] = (uint8_t) (src[4] >> 20);
    dst[29] = (uint8_t) (src[4] >> 28);
    dst[30] = (uint8_t) (src[4] >> 36);
    dst[31] = (uint8_t) (src[4] >> 44);
}

// swap (a, b) with (b, a) iff swap == 1.
void cswap(int64_t swap, field_t a, field_t b) {
    swap = -swap;
    int64_t t0 = swap & (a[0] ^ b[0]);
    a[0] ^= t0;
    b[0] ^= t0;
    int64_t t1 = swap & (a[1] ^ b[1]);
    a[1] ^= t1;
    b[1] ^= t1;
    int64_t t2 = swap & (a[2] ^ b[2]);
    a[2] ^= t2;
    b[2] ^= t2;
    int64_t t3 = swap & (a[3] ^ b[3]);
    a[3] ^= t3;
    b[3] ^= t3;
    int64_t t4 = swap & (a[4] ^ b[4]);
    a[4] ^= t4;
    b[4] ^= t4;
}

// As per RFC 7748, Chapter 5
void curve_scalar_mult(field_t dst, byte32_t n, field_t P) {
    field_t x_1; memcpy(x_1, P, sizeof(field_t));
    field_t x_2 = { 1 };
    field_t z_2 = { 0 };
    field_t x_3; memcpy(x_3, P, sizeof(field_t));
    field_t z_3 = { 1 };
    int64_t swap = 0;

    for (int32_t i = 254; i >= 0; i--) {
        int32_t bit = (n[i / 8] >> (i & 7)) & 1;
        swap ^= bit;
        cswap(swap, x_2, x_3);
        cswap(swap, z_2, z_3);
        swap = bit;

        field_t A, AA, B, BB, E, C, D, DA, CB;
        field_t t0, t1, t2, t3, t4;

        //field_add(A, x_2, z_2);
        //field_sub(B, x_2, z_2);
        field_add_sub(A, B, x_2, z_2);

        //field_add(C, x_3, z_3);
        //field_sub(D, x_3, z_3);
        field_add_sub(C, D, x_3, z_3);

        field_square(AA, A);
        field_square(BB, B);

        field_sub(E, AA, BB);

        field_mul(DA, D, A);
        field_mul(CB, C, B);

        //field_add(t0, DA, CB);
        //field_sub(t1, DA, CB);
        field_add_sub(t0, t1, DA, CB);

        field_square(x_3, t0);
        field_square(t2, t1);

        field_mul(z_3, t2, x_1);
        field_mul(x_2, AA, BB);

        field_mul_a24(t3, E);

        field_add(t4, AA, t3);
        field_mul(z_2, E, t4);
    }

    cswap(swap, x_2, x_3);
    cswap(swap, z_2, z_3);
    field_invert(z_2, z_2);         // z_2 = z_2 ^-1
    field_mul(dst, x_2, z_2);       // dst = x_2 * z_2
}

void curve25519(byte32_t dst, byte32_t n, byte32_t P) {
    byte32_t scalar;
    decode_scalar(scalar, n);

    field_t base, result;
    decode_point(base, P);
    curve_scalar_mult(result, scalar, base);

    encode_point(dst, result);
}

int crypto_scalarmult(unsigned char * dst,
                      const unsigned char *n,
                      const unsigned char *P) {
    curve25519((uint8_t *) dst, (uint8_t *) n, (uint8_t *) P);
    return 0;
}

