#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>

#include <string.h>
#include <stdio.h>

#include "../curve25519_basic.c"

#define COUNT_OF(x) (sizeof(x)/sizeof(*(x)))

static void test_c_assumptions(void **state) {
    (void) state;
    field_t x_2 = { 1 };
    assert_int_equal(x_2[0], 1);
    for (int i = 1; i < 5; i++) {
        assert_int_equal(x_2[i], 0);
    }
    assert_int_equal(sizeof(field_t), 40);
    assert_int_equal(sizeof(byte32_t), 32);
}

static void test_field_add(void **state) {
    (void) state;

    /**
    cases checked:
    - 0 + 0 = 0
    - 1 + 0 = 1
    - 1 at each byte + 0 = 1 at each byte
    - 2 at each byte + 1 at each byte = 3 at each byte
    - different prime at each byte + different prime at each byte = correct result at each byte
    - 2^25-1 + 2^25-1 = 2^26-1
    - random numbers
    */
    field_t input0[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x02, 0x02, 0x02, 0x02, 0x02},
        { 0x1F, 0x25, 0x29, 0x2B, 0x2F},
        { 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF},
        { 0x0FFFFFF, 0x0EEEEEE, 0x0DDDDDD, 0x0CCCCCC, 0x0BBBBBB}
     };

    field_t input1[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x1D, 0x17, 0x13, 0x11, 0x0D},
        { 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF},
        { 0x0111111, 0x0222222, 0x0333333, 0x0444444, 0x0555555}
   };

    field_t expected[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x03, 0x03, 0x03, 0x03, 0x03},
        { 0x3C, 0x3C, 0x3C, 0x3C, 0x3C},
        { 0x3FFFFFE, 0x3FFFFFE, 0x3FFFFFE, 0x3FFFFFE, 0x3FFFFFE},
        { 0x1111110, 0x1111110, 0x1111110, 0x1111110, 0x1111110}
    };

    field_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input0); i++) {
        field_add(actual, input0[i], input1[i]);
        assert_memory_equal(actual, expected[i], sizeof(field_t));
    }
}

static void test_field_sub(void **state) {
    (void) state;

    /**
    cases checked:
    - 0 - 0 = 0
    - 1 - 0 = 1
    - 1 at each byte - 0 = 1 at each byte
    - 2 at each byte - 1 at each byte = 1 at each byte
    - different prime at each byte - different prime at each byte = correct result at each byte
    - 2^26-1 - 2^25-1 = 2^25-1
    - random numbers
    */
    field_t input0[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x02, 0x02, 0x02, 0x02, 0x02},
        { 0x1F, 0x25, 0x29, 0x2B, 0x2F},
        { 0x3FFFFFE, 0x3FFFFFE, 0x3FFFFFE, 0x3FFFFFE, 0x3FFFFFE},
        { 0x0FFFFFF, 0x0EEEEEE, 0x0DDDDDD, 0x0CCCCCC, 0x0BBBBBB}
     };

    field_t input1[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x1D, 0x17, 0x13, 0x11, 0x0D},
        { 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF},
        { 0x0AAAAAA, 0x0999999, 0x0888888, 0x0777777, 0x0666666}
   };

    field_t expected[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x02, 0x0E, 0x16, 0x1A, 0x22},
        { 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF, 0x1FFFFFF},
        { 0x0555555, 0x0555555, 0x0555555, 0x0555555, 0x0555555}
    };

    field_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input0); i++) {
        field_sub(actual, input0[i], input1[i]);
        assert_memory_equal(actual, expected[i], sizeof(field_t));
    }
}

static void test_field_square(void **state) {
    (void) state;

    /**
    cases checked:
    - 0² = 0
    - 1² = 1
    - 2² = 4
    */
    field_t input[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x02, 0x00, 0x00, 0x00, 0x00}
    };

    field_t expected[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x04, 0x00, 0x00, 0x00, 0x00}
    };

    field_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input); i++) {
        field_square(actual, input[i]);
        assert_memory_equal(actual, expected[i], sizeof(field_t));
    }
}

static void test_integration_field_square_with_field_mul(void **state) {
    (void) state;

    /**
    cases checked:
    - (1 at each byte)² = mul(1 at each byte, 1 at each byte)
    - (2 at each byte)² = mul(2 at each byte, 2 at each byte)
    - (different prime at each byte)² = mul(different prime at each byte, different prime at each byte)
    - (random numbers)² = mul(random numbers, random numbers)
    */

    field_t input[] = {
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x02, 0x02, 0x02, 0x02, 0x02},
        { 0x02, 0x0E, 0x16, 0x1A, 0x22},
        { 0x0555555, 0x0555555, 0x0555555, 0x0555555, 0x0555555}
    };

    field_t actual_field_mul = {0};
    field_t actual_field_square = {0};

    for (size_t i = 0; i < COUNT_OF(input); i++) {
        field_mul(actual_field_mul, input[i], input[i]);
        field_square(actual_field_square, input[i]);
        assert_memory_equal(actual_field_mul, actual_field_square, sizeof(field_t));
    }
}

static void test_field_mul(void **state) {
    (void) state;

    /**
    cases checked:
    - 0 * 0 = 0
    - 1 * 0 = 1
    - 1 at each byte * 0 = 0 at each byte
    - 1 at each byte * 2 = 2 at each byte
    - 29 * 31 = 899
    */
    field_t input0[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x00, 0x00, 0x00, 0x00},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x01, 0x01, 0x01, 0x01, 0x01},
        { 0x1F, 0x00, 0x00, 0x00, 0x00},
     };

    field_t input1[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x00, 0x00, 0x00, 0x00, 0x00},
        { 0x02, 0x00, 0x00, 0x00, 0x00},
        { 0x1D, 0x00, 0x00, 0x00, 0x00},
    };

    field_t expected[] = {
        { 0x00, 0x00, 0x00, 0x00, 0x00 },
        { 0x00, 0x00, 0x00, 0x00, 0x00 },
        { 0x00, 0x00, 0x00, 0x00, 0x00 },
        { 0x02, 0x02, 0x02, 0x02, 0x02 },
        { 0x383, 0x00, 0x00, 0x00, 0x00 },
    };

    field_t actual = {0};

    for (size_t i = 3; i < COUNT_OF(input0); i++) {
        field_mul(actual, input0[i], input1[i]);
        assert_memory_equal(actual, expected[i], sizeof(field_t));
    }
}

static void test_field_invert(void **state) {
    (void) state;

    /**
    cases checked:
    - 1 => 1 (because identity)
    */
    field_t input[] = {
        { 0x01, 0x00, 0x00, 0x00, 0x00},
    };

    field_t expected[] = {
        { 0x01, 0x00, 0x00, 0x00, 0x00},
    };

    field_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input); i++) {
        field_invert(actual, input[i]);
        assert_memory_equal(actual, expected[i], sizeof(field_t));
    }
}

static void test_field_mul_a24(void **state) {
    (void) state;

    /**
    cases checked:
    - 1 => 1 (because identity)
    */
    field_t input[] = {
        { 0x01, 0x00, 0x00, 0x00, 0x00},
    };

    field_t expected[] = {
        { 0x1DB41, 0x00, 0x00, 0x00, 0x00},
    };

    field_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input); i++) {
        field_mul_a24(actual, input[i]);
        assert_memory_equal(actual, expected[i], sizeof(field_t));
    }
}


static void test_decode_scalar(void **state) {
    (void) state;

    byte32_t all_zeros = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };

    /**
    check secret key multiple of 8
    - 0 => 0
    - 3 => 0
    - 8 => 0
    - 9 => 1
    */
    uint8_t first_byte_input[] = { 0x00, 0x01, 0x07, 0x08, 0x09, 0x0D, 0xEE };
    uint8_t first_byte_expected[] = { 0x00, 0x00, 0x00, 0x08, 0x08, 0x08, 0xE8 };

    byte32_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(first_byte_input); i++) {
        byte32_t input;
        memcpy(input, all_zeros, sizeof(byte32_t));
        input[0] = first_byte_input[i];
        decode_scalar(actual, input);

        assert_int_equal(actual[0], first_byte_expected[i]);
    }

    /**
    check secret key smaller than 2^254-1
    - last byte 126 => 126
    - last byte 127 => 127
    - last byte 128 => 0 (+64)
    - last byte = 255 => 127
    */
    uint8_t last_byte_input[] = { 0x7E, 0x7F, 0x80, 0xFF };
    uint8_t last_byte_expected[] = { 0x7E, 0x7F, 0x40, 0x7F };

    for (size_t i = 0; i < COUNT_OF(last_byte_input); i++) {
        byte32_t input;
        memcpy(input, all_zeros, sizeof(byte32_t));
        input[31] = last_byte_input[i];
        decode_scalar(actual, input);

        assert_int_equal(actual[31], last_byte_expected[i]);
    }
}

static void test_decode_point(void **state) {
    (void) state;

    /**
     * cases:
     * 0 -> 0
     * FFFF.... -> 19
     * some number pattern -> some other number pattern
     * */

    byte32_t input[] = {
        {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        },
        {
            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
        }
        /*
        ,
        {
            0x12, 0x34, 0x56, 0x78, 0x9A, 0xBC, 0xDE, 0xF0,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        }
        */
    };
    int const P26 = (1 << 26) - 1;
    int const P25 = (1 << 25) - 1;
    uint64_t const P51 = ((uint64_t) 1 << 51) - 1;
    field_t expected[] = {
        { 0, 0, 0, 0, 0},
        { 0x12, 0, 0, 0},
        //{ 0xF0DEBC9A78563412 & P51, 0xF0DEBC9A78563412 >> 51, 0, 0, 0}
    };

    field_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input); i++) {
        decode_point(actual, input[i]);

        assert_memory_equal(expected[i], actual, sizeof(field_t));
    }
}

static void test_encode_point(void **state) {
    (void) state;

    /*
        ((x[0] > P26 - 19)) &
        ((x[1] & x[3] & x[5] & x[7] & x[9]) == P25) &
        ((x[2] & x[4] & x[6] & x[8]) == P26)
        ) || (x[9] > P25); */

    int const P26 = (1 << 26) - 1;
    int const P25 = (1 << 25) - 1;
    field_t input[] = {
        { 0, 0, 0, 0, 0 },
        { 0x12, 0, 0, 0 },
        //{ 0xFFFFFFFF & P26, 0xFFFFFFFF & P25, 0xFFFFFFFF & P26 }
    };

    byte32_t expected[] = {
        {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        },
        {
            0x12, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        }
        /*
        ,
        {
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 16 F => 64 of 77 bits
            0xFF, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // 3 F + 1 => +13 = 77 bits
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        }
        */
    };

    byte32_t actual = {0};

    for (size_t i = 0; i < COUNT_OF(input); i++) {
        encode_point(actual, input[i]);

        assert_memory_equal(expected[i], actual, sizeof(byte32_t));
    }
}

static void test_integration_encode_with_decode(void **state) {
    (void) state;

    byte32_t examples[] = {
        {
            0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00
        },
        {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        },
        {
            0xe6, 0xdb, 0x68, 0x67, 0x58, 0x30, 0x30, 0xdb,
            0x35, 0x94, 0xc1, 0xa4, 0x24, 0xb1, 0x5f, 0x7c,
            0x72, 0x66, 0x24, 0xec, 0x26, 0xb3, 0x35, 0x3b,
            0x10, 0xa9, 0x03, 0xa6, 0xd0, 0xab, 0x1c, 0x4c
        }
    };

    for (size_t i = 0; i < COUNT_OF(examples); i++) {
        field_t decoded;
        decode_point(decoded, examples[i]);
        byte32_t encoded;
        encode_point(encoded, decoded);
        assert_memory_equal(examples[i], encoded, sizeof(byte32_t));
    }
}

static void test_cswap(void **state) {
    (void) state;
    field_t a = { 42 };
    field_t b = { 19 };
    field_t a_ref = { 42 };
    field_t b_ref = { 19 };

    cswap(0, a, b);
    assert_memory_equal(a, a_ref, sizeof(field_t));
    assert_memory_equal(b, b_ref, sizeof(field_t));
    cswap(1, a, b);
    assert_memory_equal(a, b_ref, sizeof(field_t));
    assert_memory_equal(b, a_ref, sizeof(field_t));

    field_t a2 = { 0 };
    field_t b2 = { 42, 42, 42, 42 };
    field_t a2_ref = { 0 };
    field_t b2_ref = { 42, 42, 42, 42 };

    cswap(1, a2, b2);
    assert_memory_equal(a2, b2_ref, 32);
    assert_memory_equal(b2, a2_ref, 32);
    cswap(0, a2, b2);
    assert_memory_equal(a2, b2_ref, 32);
    assert_memory_equal(b2, a2_ref, 32);
}

// source: https://tools.ietf.org/html/rfc7748, 5.2 Test Vectors (1st)
static void test_vector_1(void **state) {
    (void) state;

    byte32_t actual = {0};
    byte32_t n = {
        0xa5, 0x46, 0xe3, 0x6b, 0xf0, 0x52, 0x7c, 0x9d,
        0x3b, 0x16, 0x15, 0x4b, 0x82, 0x46, 0x5e, 0xdd,
        0x62, 0x14, 0x4c, 0x0a, 0xc1, 0xfc, 0x5a, 0x18,
        0x50, 0x6a, 0x22, 0x44, 0xba, 0x44, 0x9a, 0xc4
    };
    byte32_t base_point = {
        0xe6, 0xdb, 0x68, 0x67, 0x58, 0x30, 0x30, 0xdb,
        0x35, 0x94, 0xc1, 0xa4, 0x24, 0xb1, 0x5f, 0x7c,
        0x72, 0x66, 0x24, 0xec, 0x26, 0xb3, 0x35, 0x3b,
        0x10, 0xa9, 0x03, 0xa6, 0xd0, 0xab, 0x1c, 0x4c
    };
    curve25519(actual, n, base_point);

    byte32_t expected = {
        0xc3, 0xda, 0x55, 0x37, 0x9d, 0xe9, 0xc6, 0x90,
        0x8e, 0x94, 0xea, 0x4d, 0xf2, 0x8d, 0x08, 0x4f,
        0x32, 0xec, 0xcf, 0x03, 0x49, 0x1c, 0x71, 0xf7,
        0x54, 0xb4, 0x07, 0x55, 0x77, 0xa2, 0x85, 0x52
    };
    assert_memory_equal(actual, expected, sizeof(byte32_t));
}

// source: https://tools.ietf.org/html/rfc7748, 5.2 Test Vectors (2nd)
static void test_vector_2(void **state) {
    (void) state;

    byte32_t actual = {0};
    byte32_t n = {
        0x4b, 0x66, 0xe9, 0xd4, 0xd1, 0xb4, 0x67, 0x3c,
        0x5a, 0xd2, 0x26, 0x91, 0x95, 0x7d, 0x6a, 0xf5,
        0xc1, 0x1b, 0x64, 0x21, 0xe0, 0xea, 0x01, 0xd4,
        0x2c, 0xa4, 0x16, 0x9e, 0x79, 0x18, 0xba, 0x0d
    };
    byte32_t base_point = {
        0xe5, 0x21, 0x0f, 0x12, 0x78, 0x68, 0x11, 0xd3,
        0xf4, 0xb7, 0x95, 0x9d, 0x05, 0x38, 0xae, 0x2c,
        0x31, 0xdb, 0xe7, 0x10, 0x6f, 0xc0, 0x3c, 0x3e,
        0xfc, 0x4c, 0xd5, 0x49, 0xc7, 0x15, 0xa4, 0x93
    };
    curve25519(actual, n, base_point);

    byte32_t expected = {
        0x95, 0xcb, 0xde, 0x94, 0x76, 0xe8, 0x90, 0x7d,
        0x7a, 0xad, 0xe4, 0x5c, 0xb4, 0xb8, 0x73, 0xf8,
        0x8b, 0x59, 0x5a, 0x68, 0x79, 0x9f, 0xa1, 0x52,
        0xe6, 0xf8, 0xf7, 0x64, 0x7a, 0xac, 0x79, 0x57
    };
    assert_memory_equal(actual, expected, sizeof(byte32_t));
}

// source: https://tools.ietf.org/html/rfc7748, 5.2 Test Vectors (2nd type of test vectors)
static void test_vector_iterations(void **state) {
    (void) state;

    byte32_t actual = {0};
    byte32_t n = {
        0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };

    byte32_t base_point = {
        0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };

    byte32_t expected_one_iteration = {
        0x42, 0x2c, 0x8e, 0x7a, 0x62, 0x27, 0xd7, 0xbc,
        0xa1, 0x35, 0x0b, 0x3e, 0x2b, 0xb7, 0x27, 0x9f,
        0x78, 0x97, 0xb8, 0x7b, 0xb6, 0x85, 0x4b, 0x78,
        0x3c, 0x60, 0xe8, 0x03, 0x11, 0xae, 0x30, 0x79
    };

    byte32_t expected_1000_iterations = {
        0x68, 0x4c, 0xf5, 0x9b, 0xa8, 0x33, 0x09, 0x55,
        0x28, 0x00, 0xef, 0x56, 0x6f, 0x2f, 0x4d, 0x3c,
        0x1c, 0x38, 0x87, 0xc4, 0x93, 0x60, 0xe3, 0x87,
        0x5f, 0x2e, 0xb9, 0x4d, 0x99, 0x53, 0x2c, 0x51
    };

#ifdef LONG_TEST
    byte32_t expected_1000000_iteration = {
        0x7c, 0x39, 0x11, 0xe0, 0xab, 0x25, 0x86, 0xfd,
        0x86, 0x44, 0x97, 0x29, 0x7e, 0x57, 0x5e, 0x6f,
        0x3b, 0xc6, 0x01, 0xc0, 0x88, 0x3c, 0x30, 0xdf,
        0x5f, 0x4d, 0xd2, 0xd2, 0x4f, 0x66, 0x54, 0x24
    };
#endif

    /**
     * For each iteration, set n to be the result of calling the function
     * and the base point to be the old value of n. The final result is the value left
     * in n.
     */
#ifdef LONG_TEST
    for (int i = 1; i <= 1000000; i++) {
#else
    for (int i = 1; i <= 1000; i++) {
#endif
        curve25519(actual, n, base_point);
        memcpy(base_point, n, sizeof(byte32_t));
        memcpy(n, actual, sizeof(byte32_t));

        if (i == 1) {
            assert_memory_equal(actual, expected_one_iteration, sizeof(byte32_t));
        }
        if (i == 1000) {
            assert_memory_equal(actual, expected_1000_iterations, sizeof(byte32_t));
        }
#ifdef LONG_TEST
        if (i == 1000000) {
            assert_memory_equal(actual, expected_1000000_iteration, sizeof(byte32_t));
        }
#endif
    }
}

int main() {
    /**
     * test naming conventions
     * test_ + name of function: handcalculated test values
     * test_integration_ + name of function: use other field operations to ensure function under test is correct
     * */
    const struct CMUnitTest tests[] = {
        cmocka_unit_test(test_field_mul),
        cmocka_unit_test(test_c_assumptions),
        cmocka_unit_test(test_cswap),
        cmocka_unit_test(test_field_add),
        cmocka_unit_test(test_field_sub),
        cmocka_unit_test(test_field_square),
        cmocka_unit_test(test_integration_field_square_with_field_mul),
        cmocka_unit_test(test_field_invert),
        cmocka_unit_test(test_field_mul_a24),
        cmocka_unit_test(test_decode_scalar),
        cmocka_unit_test(test_decode_point),
        cmocka_unit_test(test_encode_point),
        cmocka_unit_test(test_integration_encode_with_decode),
        cmocka_unit_test(test_vector_iterations),
        cmocka_unit_test(test_vector_1),
        cmocka_unit_test(test_vector_2)
    };

    return cmocka_run_group_tests(tests, NULL, NULL);
}
